#
# VT50-2000-2016-84-28-4
# $Copyright: Copyright (c) 2016 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $RCSfile: OraDGBrokerAgent.pm $ $Revision: 817879d93479bf83c80cc348a81c3cd63f5cd9d0 $
#
#highest message no. used: 1006

package OraDGBrokerAgent;
use ArrayRep;
@ISA = ("ArrayRep");

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);
use ag_i18n_inc;

my $SU = "/bin/su";
my $PASSWDFILE = "/etc/passwd";

@EXPORT = qw(
);

$VERSION = '1.00';

require 5.000;

# Environment type
use constant ENV_SI => 0;
use constant ENV_RAC => 1;
use constant ENV_UNKNOWN => -1;

%default = {
	resname => '',

	VADCtlMode => '',

	# Elements to store the resource atributes
	# ========================================

	# Corresponding(linked) Oracle resource name
	LinkRes => '',

	# Oracle instance ID used as OARCLE_SID when calling the dgmgrl
	Sid => '',

	# Oracle user ID used to SU for dgmgrl
	Owner => '',

	# Oracle home directory used for ORACLE_HOME
	Home => '',

	# Force failover even if remote cluster is still active but
	# remote RAC instance (service group) is offline
	ForceDB => '',

	# AgentDebug will enable exhaustive messages from every 
	# data guard broker CLI call
	AgentDebug => '',

	# Elements holding static information
	# ==============================================

	# The hostname as retrieved from the OS and used by Oracle
	hostname => '',

	# The nodename as used by VCS/VAD
	nodename => '',

	# The service group name for this resource
	grpname => '',

	# The name of the remote cluster
	remcls => '',

	# The Oracle command to call the data guard broker CLI
	dgmgrl => '',

	# Timeout / alarm value for the dgmgrl command default 20 sec
	# This does not apply to the failover / switchover calls!
	dgalarm => '',

	# Retry count for the dgmgrl show commands
	dgsretry => '',

	# Elements holding dynamic data
	#==============================

	# The state of the remote cluster
	remstate => '',

	# stdout lines from external commands called using backticks
	cmdout => '',

	# The unique name of the database as returned by dgmgrl
	# for the instance for RAC. For SI its DBName attribute value
	dbname => '',

	# The current intended role of the instance as reported by dgmgrl
	dbrole => '',

	# The errorr message from the last dgmgrl command
	orclerrmsg => '',

	# Indicates if it is SI or RAC env.
	env_type  => '',
	
	#Database user name
	dbuser => '',

	#database password
	dbpassword => ''
};

#
# Fill the global hash with all resource relevant information including
# values retrieved from the attribute list
#
# Return values:	pointer to global hash = success
# 			undef = something went wrong an the calling
# 			EP has to exit / return state unknown
sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;

	my $ret;
	my $self = { %default };

	$self->{'resname'} = shift;
	my @Args = @_;

	# Check for VAD controlmode return value if > 0
	my $clus_prod = $ENV{"CLUSTER_PRODUCT"};
	$clus_prod = "VCS" if (! defined($clus_prod));
	$self->{'VADCtlMode'} = 0;
	if ($clus_prod eq "VAD") {
		($ret, $self->{'VADCtlMode'}) = VCSAG_GET_ATTR_VALUE ("ControlMode", -1, 1, @ARGV);
		return undef if ($ret != $VCSAG_SUCCESS);
	}

	# Get Arguments

	($ret, $self->{'LinkRes'}) = VCSAG_GET_ATTR_VALUE ("LinkRes", -1, 1, @Args);
	return undef if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'Sid'}) = VCSAG_GET_ATTR_VALUE ("Sid", -1, 1, @Args);
	return undef if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'dbname'}) = VCSAG_GET_ATTR_VALUE ("DBName", -1, 1, @Args);
	return undef if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'Owner'}) = VCSAG_GET_ATTR_VALUE ("Owner", -1, 1, @Args);
	return undef if ($ret != $VCSAG_SUCCESS);

	my $fnd = 0;
	open(PWF, "<", $PASSWDFILE);
	while (<PWF>) {
		next unless /^$self->{'Owner'}:/;
		$fnd = 1;
		last;
	}
	close PWF;
	unless ($fnd) {
		VCSAG_LOG_MSG("E", "User defined by attribute Owner does not exist in local password file", 501);
		return undef;
	}

	($ret, $self->{'Home'}) = VCSAG_GET_ATTR_VALUE ("Home", -1, 1, @Args);
	return undef if ($ret != $VCSAG_SUCCESS);


	($ret, $self->{'ForceDB'}) = VCSAG_GET_ATTR_VALUE ("ForceDB", -1, 1, @Args);
	if ($ret != $VCSAG_SUCCESS) {
		# ForceDB was added for Q4/2009 it may not be in the argument list
		$self->{'ForceDB'} = '';
	}

	($ret, $self->{'AgentDebug'}) = VCSAG_GET_ATTR_VALUE ("AgentDebug", -1, 1, @Args);
	return undef if ($ret != $VCSAG_SUCCESS);

	($ret, $self->{'Encoding'}) = VCSAG_GET_ATTR_VALUE ("Encoding", -1, 1, @Args);
	return undef if ($ret != $VCSAG_SUCCESS);
	
	($ret, $self->{'dbuser'}) = VCSAG_GET_ATTR_VALUE ("DBUserName", -1, 1, @Args);
	#return undef if ($ret != $VCSAG_SUCCESS);
	if ($self->{'dbuser'} ne "") {
		#DBUserName is configured.
		($ret, $self->{'dbpassword'}) = VCSAG_GET_ATTR_VALUE ("DBPassword", -1, 1, @Args);
		if (($ret != $VCSAG_SUCCESS) or ($self->{'dbpassword'} eq "")) {
			VCSAG_LOG_MSG("E", "Attribute DBUserName and DBPassword both should be configured together.", 1003);
			return undef;
		}
	} else {
		 #DBUserName is not configured.
		($ret, $self->{'dbpassword'}) = VCSAG_GET_ATTR_VALUE ("DBPassword", -1, 1, @Args);
		if (($ret = $VCSAG_SUCCESS) or ($self->{'dbpassword'} ne  "")) {
                	VCSAG_LOG_MSG("E", "Attribute DBUserName and DBPassword both should be configured together.", 1003);
                	return undef;
		}
	}
	
	chomp($self->{'hostname'} = `uname -n`);

	$self->{'dgmgrl'} = "(ORACLE_SID=$self->{'Sid'}\;";
	$self->{'dgmgrl'} .= "ORACLE_HOME=$self->{'Home'}\;";
	$self->{'dgmgrl'} .= "export ORACLE_SID ORACLE_HOME\;";
	$self->{'dgmgrl'} .= "$SU $self->{'Owner'} -c \"";

	if (($self->{'dbuser'} eq "") or ($self->{'dbpassword'} eq "")) {
		$self->{'dgmgrl'} .= "$self->{'Home'}/bin/dgmgrl / \\\"<DGMCMD>\\\"\")";
	} else {
		my $Pword = `$HADECRYPT $self->{'dbpassword'}`;
                if ($? != 0) {
                        VCSAG_LOG_MSG("W", "Encountered errors while decrypting password!", 1002);
                        $Pword = "";
			 return undef;
                }

		$self->{'dgmgrl'} .= "$self->{'Home'}/bin/dgmgrl $self->{'dbuser'}/$Pword \\\"<DGMCMD>\\\"\")";
	}

	if ($self->{'ForceDB'}) {
		$self->{'srvctl'} = "(ORACLE_HOME=$self->{'Home'}\;";
		$self->{'srvctl'} .= "export ORACLE_HOME\;";
		$self->{'srvctl'} .= "$self->{'Home'}/bin/srvctl status";
		$self->{'srvctl'} .= " database -d $self->{'ForceDB'})";
	}
	$self->{'dgalarm'} = 60;
	$self->{'dgsretry'} = 5;

        $self->{'env_type'} = ENV_UNKNOWN;
        my ($sg)  = split( '\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Group` );
        my ($pll) = split( '\n', `$ag_i18n_inc::HAGRP -value $sg Parallel` );
  
      if ( $pll == 1 ) {
            $self->{'env_type'} = ENV_RAC;
        } elsif ( $pll == 0 ) {
            $self->{'env_type'} = ENV_SI;
        } else {
            $self->{'env_type'} = ENV_UNKNOWN;
        } 

	if($self->{'env_type'} == ENV_SI && $self->{'dbname'} eq '') {
		VCSAG_LOG_MSG("E", "Please specify unique database name.", 533);
		return undef;
	}
	
	bless $self, $class;

	$self->get_local_nodename;

	return $self;
}

sub destroy {}

# Flush all child groups on the local node to reset faults of
# Oracle resources caused by Broker interaction
sub cg_flush {
	my $self = shift;

	my $cgrp = '';

	open(GDEP, "$ag_i18n_inc::HAGRP -dep $self->{'grpname'} 2>&1 |") or
		return 1;
	while (<GDEP>) {
		next if /^#/;
		(undef, $cgrp, undef) = split;
		system("$ag_i18n_inc::HAGRP -flush $cgrp -sys $self->{'nodename'} -localclus 2>&1 1>/dev/null");
	}
	close GDEP;
	return 0;
}

# Freeze / unfreeze all child groups to supress cluster interaction
# which contain a resource of type Oracle for the same Sid
# while the broker restarts instances during a switch / failover
#
# Return value 	0 - success for all groups
#		1 - error on open $HAGRP -dep
# 		>0 the accumalated return values of $HAGRP -[un]freeze
sub cg_freeze {
	my $self = shift;
	my $froper = shift;
	if ($froper) {
		$froper = "-freeze";
	}
	else {
		$froper = "-unfreeze";
	}
	my $cgrp = '';
	my $ret = 0;

	open(GDEP, "$ag_i18n_inc::HAGRP -dep $self->{'grpname'} 2>&1 |") or
		return 1;
	while (<GDEP>) {
		next if /^#/;
		(undef, $cgrp, undef) = split;
		my $gres = `$ag_i18n_inc::HAGRP -resources $cgrp 2>&1`;
		chomp $gres;
		foreach my $indres (split(/\n/, $gres)) {
			next unless (`$ag_i18n_inc::HARES -value $indres Type 2>&1` =~ /Oracle/);
			next unless (`$ag_i18n_inc::HARES -value $indres Sid 2>&1` =~ /$self->{'Sid'}/);
			$ret += system("$ag_i18n_inc::HAGRP $froper $cgrp -localclus 2>&1 1>/dev/null");
			last;
		}
	}
	close GDEP;
	return $ret;
}

# Clear any oracle SG faults
sub cg_clear {
	my $self = shift;

      	my ($sg)  = split( '\n', `$ag_i18n_inc::HARES -value $self->{'LinkRes'} Group` );
      	my ($state) = split( '\n', `$ag_i18n_inc::HAGRP -state $sg -sys $self->{'nodename'}` );

	if($state =~ /FAULTED/) {
		system("$ag_i18n_inc::HAGRP -clear $sg 2>&1 1>/dev/null");
	}
	return 0;
}

# Probe oracle resource
sub res_probe {
	my $self = shift;

	system("$ag_i18n_inc::HARES -probe $self->{'LinkRes'} -sys $self->{'nodename'} 2>&1 1>/dev/null");

	return 0;
}

# Execute a single data guard CLI command given as the second value in the
# call list
# 
# Return values:	the return value of the dgmgrl command plus
# 			stdout / stderr in $self->{'cmdout'}
sub exec_dgmgrl {
	my $self = shift;
	my $dgbcmd = shift;

	# Reset the cdmout variable since it is extended
	$self->{'cmdout'} = '';

	my $cmd = $self->{'dgmgrl'};
	$cmd =~ s/<DGMCMD>/$dgbcmd/;
 
	VCSAG_LOGDBG_MSG(1, "exec_dgmgrl running dgmgrl command :: $dgbcmd ");
	my $pid = open(DGCM, "$cmd 2>&1 |") or return 2;

	eval {
		local $SIG{ALRM} = sub { die "dgalarm timer exhausted" };
		alarm $self->{'dgalarm'};
		while (<DGCM>) {
			$self->{'cmdout'} .= $_;
		}
		close DGCM;
		alarm 0;
	};

	if ($@ and $@ =~ /dgalarm timer exhausted/) {
		kill 9, $pid;
		close DGCM;
		VCSAG_LOG_MSG("W", "dgalarm timer exhausted when calling $cmd", 523, $cmd) if $self->{'AgentDebug'};
		$self->{'cmdout'} .= "VCS $self->{'resname'} dgalarm timer = $self->{'dgalarm'} sec exhausted while executing dgmgrl command";
		return 1;
	}
	return 0;
}

# Execute a Cluster or Data Guard Broker command at a cluster site
# using an action EP
#
# Return values:	0 on sucessful execution on at least one node
# 				$self->{'cmdout'} will contain the action or
# 				command output
# 			1 - if execution does not complete on any remote node
sub exec_action {
	my $self = shift;
	my $action_args = shift;
	my $nodename = shift;
	my $clusname = shift;
	my $out = "";

	my ($clnode, $nname, $res_state, $rem_clus);

	if($self->{'env_type'} == ENV_RAC) {
		open(RNOD, "$ag_i18n_inc::HARES -state $self->{'resname'} 2>&1 |") or
			return 1;

		# Call the ActRemote action entry point with the rest of
		# action_args as a string against each remote node
		while (<RNOD>) {
			next if /^#/;
			next if /\s+localclus\:/;
			(undef, undef, $clnode, undef) = split;
			(undef, $nname) = split(/:/, $clnode);
			next unless (defined $nname);
			$self->{'cmdout'} = `$ag_i18n_inc::HARES -action $self->{'resname'} ActRemote -actionargs $action_args -sys $nname -clus $self->{'remcls'} 2>&1`;
			return 0 unless $?;
		}
	}

	if($self->{'env_type'} == ENV_SI) {

		if($nodename ne '' && $clusname ne '') {
			my $actioncmd = "$ag_i18n_inc::HARES -action $self->{'resname'} ActRemote -actionargs $action_args -sys $nodename -clus $clusname";
			VCSAG_LOG_MSG("I", "Executing action [$actioncmd]", 533, $actioncmd);
			$self->{'cmdout'} = `$actioncmd 2>&1`;
			return 0 unless $?;
		} else {
			open(RNOD, "$ag_i18n_inc::HARES -state $self->{'resname'} 2>&1 |") or
				return 1;
			

			# Call the ActRemote action entry point with the rest of
			# action_args as a string against remote node where oracle is online
			while (<RNOD>) {
				next if /^#/;
				next if /\s+localclus\:/;
				#next if /\sOFFLINE/;
				(undef, undef, $clnode, $res_state) = split;
				($rem_clus, $nname) = split(/:/, $clnode);
				next unless (defined $nname);
				my $actioncmd = "$ag_i18n_inc::HARES -action $self->{'resname'} ActRemote -actionargs $action_args -sys $nname -clus $rem_clus";
				VCSAG_LOG_MSG("I", "Executing action [$actioncmd]", 533, $actioncmd);
				$self->{'cmdout'} = `$actioncmd 2>&1`;
				my $ret = $?;
				if (($ret != 0) && ($self->{'cmdout'}  =~ /is frozen/)) {
					VCSAG_LOG_MSG("E", "Remote Oracle dataguard service group is frozen", 600);
					return 1;
				}
				#if return status is 2 , means oracle resource is not online at this node so check other nodes
				next if $ret == 2; 
				return 0 if $ret == 0;
				if( $ret == 1) {   
					VCSAG_LOG_MSG("E", "Check the action arguments", 1004);
					return 1;
				}
				return 1 if $ret == 3; 
			}
			VCSAG_LOG_MSG("W", "Unable to find any node in the cluster which have online oracle resource", 1005);
			return 1;
		}
	}
}

# Function to extend the timout value of an entry point
# useful for action and info with short timeout values
sub extend_ep_timeout {
	my $self = shift;
	my $tmo_val = shift;
	return 1 unless (defined $tmo_val);

	my $tmo_ext_file = "/var/VRTSvcs/log/tmp/" . "$self->{'resname'}" . ".tmo";
	open(TEF, "> $tmo_ext_file") or return 1;
	print TEF "$tmo_val";
	close TEF;
	return 0;
}

# Get group and remote cluster name for the resource
#
# Return values:	0 - success, $self->{'grpname'} & $self->{'remcls'}
#			1 - fault in retrieving information
sub get_clus_info {
	my $self = shift;

	$self->{'grpname'} = `$ag_i18n_inc::HARES -value $self->{'resname'} Group 2>&1`;
	return 1 if ($?);

	chomp ($self->{'grpname'});
	open(RCLUS, "$ag_i18n_inc::HAGRP -display $self->{'grpname'} -attribute Cluster 2>&1 |");
	while (<RCLUS>) {
		next if /^#/;
		chomp;
		(undef, undef, undef, $self->{'remcls'}) = split;
		last;
	}
	close RCLUS;
	return 1 if ($self->{'remcls'} eq '');

	return 0;
}

# Retrieve the name of the database associated by Oracle with the instance
# defined by the SID attribute
#
# Return values:	0 - success, $self->{'dbname'} is updated
# 			1 - database name can not be retrieved
sub get_db_name {
	my $self = shift;

	if(($self->{'dbname'} eq '') &&  ($self->{'env_type'} == ENV_RAC)) {
		my $dgbcmd = "show instance '".$self->{'Sid'}."'";
	
		$self->exec_dgmgrl($dgbcmd);
		$self->{'dbname'} = '';

		foreach (split(/\n/, $self->{'cmdout'})) {
			chomp;
			if (s/^instance.*$self->{'Sid'}.*database\s+\'//i) {
				s/\'.*$//;
				$self->{'dbname'} = $_;
				next;
			}
			if (/ORA-\d+/) {
				$self->{'orclerrmsg'} = $_;
				VCSAG_LOG_MSG("W", "The show instance call returned Oracle message $self->{'orclerrmsg'}", 520, $self->{'orclerrmsg'}) if $self->{'AgentDebug'};
				last;
			}
			last if (/SUCCESS/i);
		}
	}
	if ($self->{'dbname'} ne '') {
		return 0;
	}
	else {
		VCSAG_LOG_MSG("E", "Can not determine database name for instance $self->{'SID'}", 521, $self->{'SID'}) if $self->{'AgentDebug'};
		return 1;
	}
}

# Get the value of the "Role" field from the dgmgrl "show database" command
#
# Return values:	0 - success, $self->{'dbrole'} is updated
#			1 - the current role could not be retrieved
#			    for example Oracle might not be running
#
#			extensive message if AgentDebug is enabled
sub get_db_role {
	my $self = shift;
	my $instmatch = 0;

	my $dgbcmd = "show database '".$self->{'dbname'}."'";
	$self->exec_dgmgrl($dgbcmd);

	$self->{'dbrole'} = '';

	foreach (split(/\n/, $self->{'cmdout'})) {
		chomp;
		if (s/^\s+Role:\s+//i) {
			if (/PRIMARY/i) {
				$self->{'dbrole'} = $_;
			}
			else {
				$self->{'dbrole'} = $_;
			}
		}
		if($self->{'env_type'} == ENV_RAC) {
			if (s/\s+$self->{'Sid'}//i) {
				$instmatch = 1;
				if (/apply instance/) {
					$self->{'dbrole'} .= " apply instance";
				}
			}
		}
		if (/ORA-\d+/) {
			$self->{'orclerrmsg'} = $_;
			VCSAG_LOG_MSG("W", "The show instance call returned Oracle message $self->{'orclerrmsg'}\n Output : $self->{'cmdout'}", 520, $self->{'orclerrmsg'}, $self->{'cmdout'}) if $self->{'AgentDebug'};
			last;
		}
		last if (/SUCCESS/i);
	}
	if ($self->{'dbrole'} eq '' || ($self->{'env_type'} == ENV_RAC && $instmatch == 0)) {
		VCSAG_LOG_MSG("E", "Can not determine role for database $self->{'dbname'}, $dgbcmd returned $self->{'cmdout'}", 522, $self->{'dbname'}, $dgbcmd,  $self->{'cmdout'}) if $self->{'AgentDebug'};
		return 1;
	}
	return 0;
}

# Function : get_local_nodename
# Desc     : Retrieve the name of the local node
# Returns  : 0 success - $self->{'nodename'} is updated

sub get_local_nodename {
	my $self = shift;
	my $SysName = "${VCSCONF}/conf/sysname";
	my $GlobZoneName = "$VCSCONF/.vcshost";
	my $LocSys;

	if ( -r "$SysName" ) {
		open(SNAME, "<$SysName");
	}
	elsif ( -r "$GlobZoneName" ) {
		open(SNAME, "<$GlobZoneName");
	}
	else {
		open(SNAME, "uname -n |");
	}
	chomp($self->{'nodename'} = <SNAME>);
	close SNAME;
	return 0;
}

# Get the state of the remote cluster for the resource.
# Only a two cluster = single standby configuration is supported!
#
# Return values:	0 - success, $self->{'remstate'} is updated
# 			1 - remote state can not be retrieved

sub get_remote_state {
	my $self = shift;

	$self->{'remstate'} = `$ag_i18n_inc::HACLUS -state $self->{'remcls'} 2>&1`;
	chomp($self->{'remstate'});

	return 0;
}

# Get the state of the remote RAC instances
#
# Return values:	0 - success, $self->{'cmdout'} is updated
# 			1 - remote state can not be retrieved

sub rac_state {
	my $self = shift;

	$self->{'cmdout'} = `$self->{'srvctl'} 2>&1`;
	print $self->{'cmdout'};

	return 0;
}


# Function : get_oracle_version
# Desc     : It is used to find the version of Oracle.
# Input    : $version - passed as reference. Oracle version will be copied.
#          : $release - passed as reference. Oracle release will be copied.
# Returns  : always 1
#
sub get_oracle_version
{
        my $self = shift;
	my ($version, $release) = @_;
        my ($out, @out1);
        my ($lineval, $string, $str);
        my $Home = $self->{'Home'};
	
	$ENV{'ORACLE_HOME'}=$Home;
        $out = `$self->{'Home'}/bin/sqlplus -V`;
        @out1 = split(/\n/,$out);
        foreach $lineval (@out1) {
                chomp $lineval;
                if($lineval=~/Release/) {
                        $string = $lineval;
                        ($str = $string) =~ s/.*Release\s+(\d+)\.(\d+).*/$1/;
                        $release = $2;
                        last;
                }
        }

        $_[0] = $str;
        $_[1] = $release;

        return 1;
}

# Get remote primary database name in Single Instance environment
#
# Return values:	0 - success, $self->{'cmdout'} is updated
# 			1 - remote primary database name can not be retrieved

sub get_remote_primary_db_name {
	my $self = shift;

	my $dgbcmd = "show configuration";
	$self->exec_dgmgrl($dgbcmd);

	my $remote_db_name = '';
	my $dbstart = 0;

	foreach (split(/\n/, $self->{'cmdout'})) {
		chomp;
		$_ =~ s/^\s+//;
		if ($_=~/Primary/) {
			($remote_db_name, undef, undef) = split;
			if($remote_db_name ne $self->{'dbname'}) {
				last;
			}
			else {
				next;
			}
		}
		if (($remote_db_name eq '') && /ORA-\d+/) {
			$self->{'orclerrmsg'} = $_;
			VCSAG_LOG_MSG("W", "The show configuration call returned Oracle message $self->{'orclerrmsg'}", 534, $self->{'orclerrmsg'}) if $self->{'AgentDebug'};
			last;
		}
		last if (/SUCCESS/i);
	}
	if ($remote_db_name eq ''){
		VCSAG_LOG_MSG("E", "Can not determine name of the remote primary database for $self->{'dbname'}, $dgbcmd returned $self->{'cmdout'}", 535, $self->{'dbname'}, $dgbcmd,  $self->{'cmdout'});
	}

	return $remote_db_name;
}

# Find out if remote primary database is reachable or not in Single Instance environment
#
# Return values:	0 - primary is not reachable
# 			1 - primary is reachable

sub is_primary_reachable {
	my $self = shift;

	my $pri_db = $self->get_remote_primary_db_name();

	if($pri_db eq '') {
		return 0;
	}

	my $pri_unreachable = 0;

	my $dgbcmd = "show database '".$pri_db."'";
	$self->exec_dgmgrl($dgbcmd);

	foreach (split(/\n/, $self->{'cmdout'})) {
		chomp;
		if (/DGM-17016/ || /ORA-12514/ || /ORA-16625/ ) {
			$self->{'orclerrmsg'} = $_;
			VCSAG_LOG_MSG("W", "The show database call returned Oracle message $self->{'orclerrmsg'}", 536, $self->{'orclerrmsg'}) if $self->{'AgentDebug'};
			$pri_unreachable = 1;
			last;
		}
		last if (/SUCCESS/i);
	}
	if ($pri_unreachable) {
		return 0;
	}
	return 1;
}

#
# Get the nodename where Oracle SG is online (for SI env)
#

sub get_active_node {
	my $self = shift;
	my $active_node = '';
	my $res_state = '';
	my $clusname = '';

      	my @clusnames  = split( '\n', `$ag_i18n_inc::HACLUS -value ClusterName` );
	$clusname = $clusnames[0];

	open(RNOD, "$ag_i18n_inc::HARES -state $self->{'LinkRes'} 2>&1 |") or
		return 1;

	while (<RNOD>) {
		next if /^#/;
		(undef, undef, $active_node, $res_state) = split;
		if($res_state =~ "ONLINE") {
			#output of the ActRemote EP that will be used
			print "$active_node:$clusname\n";
			return 0;
		}
	}

	print "$active_node:$clusname\n";

	#if there was no node in the cluster where Oracle SG was online, return the last node that we found.
	return 1;
}

#Find out if database needed recovery or not in Single Instance environment
#
#Return Values:		0 - recovery not needed
#			1 - recovery needed

sub Is_recovery_needed {
	my $self = shift;
	my $recoverdb = @_;
	my $dgbcmd = "show configuration";
	my ($version , $release);

	$self->get_oracle_version($version, $release);
	my $oracle_version = $version.".".$release;

	$self->exec_dgmgrl($dgbcmd);
	my @lines = split(/\n/, $self->{'cmdout'});
	foreach(@lines) {
		$_ =~ s/^\s+//;
		if ($_=~/Physical standby/) {
			($recoverdb, undef, undef) = split;
			VCSAG_LOGDBG_MSG(1, "Recover db: $recoverdb.");
			$_[0] = $recoverdb; # for returning the recovery needed database name

			#check if it is oracle 11gR1 and database need recovery
			if($oracle_version eq "11.1" and $_ =~/\(disabled\)/) {
				my $ret = $self->exec_action("DGMGRL ".$recoverdb." ".$dgbcmd);
				my @out1 = split(/\n/ , $self->{'cmdout'});
				foreach(@out1) {
					chomp;
					if(/ORA-16795/ or /standby database needs to be re-created/i) {
						return 1; # recovery needed
					}
				}
			}
		}
		if (/^ORA-16661: the standby database needs to be reinstated/i) {
			return 1;
		}
	}
	return 0; # recovery not needed
}

#Get the remote database name whose instance name is provided as parameter
#

sub get_instance_db_name {
	my $self = shift;
	my $instance_name = shift;
	my $rem_db_name = '';	

	my $ret = $self->exec_dgmgrl("show instance ".$instance_name);
	if($ret == 0) {
		foreach(split(/\n/ , $self->{'cmdout'})) {
			next if $_ =~ /^\s+/;
			if(s/^Instance//m) {
				my $tmp = $_;
				$tmp =~ s/.*database \'(.*?)\'/$1/m;
				$rem_db_name = $1;
				last;
			}
		}
	}

	return $rem_db_name;	
}

# Perform role change operation in Single Instance environment
#
# Return values:	0 - success, $self->{'cmdout'} is updated

sub SI_failover {
	my $self = shift;
	my $ret = 0;

	#If Primary is up and reachable it is planned migration(switchover)	
	if($self->is_primary_reachable()) {
		#switchover

		$self->cg_freeze(1);

		# Freeze remote child groups before starting switchover
		$ret = $self->exec_action("CGfreeze");
		if($ret != 0) {
			VCSAG_LOG_MSG("E", "Failed to freeze remote service group. Please check remote cluster.", 601);
			return $ret;
		}
	
		$ret = $self->exec_dgmgrl("switchover to \'$self->{'dbname'}\'");
		VCSAG_LOG_MSG("I", "Switchover returned exit status $ret \n Output of the operation :: \n $self->{'cmdout'}", 518, $ret, $self->{'cmdout'});

		# Check output for additional Broker directives
		my $oper_flag = 0;
		my $oper_flag_11gR1 = 0;
		my ($version,$release);
		$self->get_oracle_version($version , $release);
		my $oracle_version = $version.".".$release;
		my @lines = split(/\n/, $self->{'cmdout'});
		foreach (@lines) {
			chomp;
			if (/Error:/) {
				$self->{'orclerrmsg'} = $_;
				VCSAG_LOG_MSG("E", "The switchover command failed with Oracle message $self->{'orclerrmsg'}", 545, $self->{'orclerrmsg'});
				$ret = 1;
			}
			
			if($_ =~ /^ORA-01031/ and $oracle_version eq "11.1") {
		        	$oper_flag_11gR1 = 1;
				next;
			}

			if($oper_flag_11gR1) {
				next unless $_ =~ /^You must /i;
				$_ =~ s/^You must //i;
				$oper_flag = 1;
			}

			if (/^Please complete the following steps /i) {
				$oper_flag = 1;
				next;
			}

			if($oper_flag) {
				my $oper = '';
				$_ =~ s/^\s+//;
				if (/^shut *down /i) {
					$oper = "shutdown immediate";
				}
				if(/^start *up /i or ($oracle_version eq "11.1" and /^start /i)) {
					$oper = "startup";
				}
				if ($oper eq '') {
					VCSAG_LOG_MSG("W", "Unknown operation request from the broker exit:\n$_", 532, $_);
					next;
				}
				my $act_inst = $_;
				my $rem_instance = '';
				my $rem_db = '';
				$act_inst =~ s/\"(.*?)\"/$1/m;
				$rem_instance = $1;
				if($oracle_version eq "11.1") {
					$rem_db = $self->get_instance_db_name($rem_instance);
				}
				else {
					$act_inst =~ s/\"(.*?)\"/$1/m;
					$rem_db = $1;
				}
				VCSAG_LOGDBG_MSG(1, "The remote database name is $rem_db and remote instance $rem_instance");
				if ($rem_db eq $self->{'dbname'}) {
					# This is the local instance
					$ret = $self->exec_dgmgrl("$oper");
				}
				else {
					VCSAG_LOGDBG_MSG(1, "Executing remote action EP for $1");
					$ret = $self->exec_action("DGMGRL ".$rem_instance." ".$oper);
					if($ret != 0) {
                			        VCSAG_LOG_MSG("W", "Successfully changed Database role but Failed to restart remote database $rem_db and remote instance $rem_instance. Please restart remote instance.", 602);
                        			$ret = 0;
                			}
				}
			}
		}

		#If it was needed to restart remote db, clear sg faults
		$self->exec_action("CGclear");

		#If it was needed to restart remote db, probe remote oracle resource
		$self->exec_action("Resprobe");

		#unfreeze
		$self->cg_freeze(0);
		$self->exec_action("CGunfreeze");

	}else {
		#failover
		$ret = $self->exec_dgmgrl("failover to \'$self->{'dbname'}\'");
		VCSAG_LOG_MSG("I", "Failover returned exit status $ret \n Output of the operation :: \n $self->{'cmdout'}", 517, $ret, $self->{'cmdout'});

		my @lines = split(/\n/, $self->{'cmdout'});
		foreach (@lines) {
			chomp;
			if (/Error:/) {
				$self->{'orclerrmsg'} = $_;
				VCSAG_LOG_MSG("E", "The failover command failed with Oracle message $self->{'orclerrmsg'}", 546, $self->{'orclerrmsg'});
				$ret = 1;
			}
		}
	}

	return $ret if($ret);

	# Wait until the dust settles, if necessary until online timeout
	while (1) {
		$self->get_db_role;
		last if (($self->{'dbrole'} =~ /PRIMARY/i) &&
			($self->{'orclerrmsg'} !~ /ORA-16610.*[SWITCH|FAIL]OVER to $self->{'dbname'}.*in progress/i));
		sleep 20;
	}

	return $ret;
}

# Perform Flashback Recovery (after failover)
#
# Return values:	0 - success, $self->{'cmdout'} is updated

sub flashback_recover {
	my $self = shift;
	my $recoverdb = shift;
	my $ret = 0;
	my $active_remote_node = '';
	my $clusname = '';

	#get the active node where Oracle is running, so that we execute all remote actions there only
	if($self->{'env_type'} == OraDGBrokerAgent::ENV_SI) {
		$ret = $self->exec_action("ActiveNode");
		my @lines = split(/\n/, $self->{'cmdout'});
		foreach my $line (@lines) {
			next if ($line =~ "Output is:");
			($active_remote_node, $clusname) = split(":", $line);	
			last if ($active_remote_node ne '' || $clusname ne '');
		}
		VCSAG_LOGDBG_MSG(1, "flashback_recover :: Active Remote Node: $active_remote_node Cluster name: $clusname");
		if($active_remote_node eq '' || $clusname eq '') {
			VCSAG_LOG_MSG("E", "Could not find the remote cluster or a system in remote cluster where Oracle is currently ONLINE. Administrative intervention required.", 546);
			print "Could not find the remote cluster or a system in remote cluster where Oracle is currently ONLINE. Administrative intervention required.\n";
			return 1;
		}
	}

	# Freeze remote child groups 
	$ret = $self->exec_action("CGfreeze");

	#get instance
	my $binst = 0;
	my $rem_instance = '';
	$ret = $self->exec_dgmgrl("show database $recoverdb");

	my @lines = split(/\n/, $self->{'cmdout'});
	foreach (@lines) {
		$_ =~ s/^\s+//;
		VCSAG_LOGDBG_MSG(1, "Recover db show database line: $_");
		if ($binst) {
			$rem_instance = $_;
			VCSAG_LOGDBG_MSG(1, "Recover db instance: $rem_instance");
			last;
		}


		if ($_ =~ /Instance/i) {
			$binst = 1;
		}
	}


	#From the unit-testing and related posts from the internet, the reinstate command often fails when executed first time
	#but succeeds in next attempt. It is also observed that it is not sufficitent to just retry the reinstate command from 
	# the agent. But the whole procedure needs to be retried.

	my $retryCnt = 3;
	my $retryFlag = 0;
	do {
		$retryFlag = 0;

		my $oper = "shutdown immediate";
		$ret = $self->exec_action("DGMGRL " . $rem_instance . " " . $oper, $active_remote_node, $clusname);
		if($ret) {
			VCSAG_LOG_MSG("E", "There was an error executing $oper. Administrative intervention required.", 543, $ret);
			print "There was an error executing $oper. Administrative intervention required.\n";
			return 1;
		}

		$oper = "startup mount";
		$ret = $self->exec_action("DGMGRL " . $rem_instance . " " . $oper, $active_remote_node, $clusname);
		if($ret) {
			VCSAG_LOG_MSG("E", "There was an error executing $oper. Administrative intervention required.", 543, $ret);
			print "There was an error executing $oper. Administrative intervention required.\n";
			return 1;
		}

		#give some time to db to come up to avoid ORA-16653 error for reinstate command
		# Wait until the recover database starts (until show configuration returns ORA-16795: the standby database needs to be re-created)
		my $breaktheloop = 0;
		while (1) {
			$oper = "show configuration";
			$ret = $self->exec_action("DGMGRL " . $rem_instance . " " . $oper, $active_remote_node, $clusname);
			my @lines = split(/\n/, $self->{'cmdout'});
			foreach (@lines) {
				chomp;
				if (/ORA-16795/i) {
					$self->{'orclerrmsg'} = $_;
					VCSAG_LOGDBG_MSG(1, "flashback_recover :: Show configuration returned Oracle message $self->{'orclerrmsg'}");
					$breaktheloop = 1;
					last;
				}
			}
			sleep 20;
			last if($breaktheloop);
		}

		VCSAG_LOGDBG_MSG(1, "flashback_recover :: Executing reinstate command now");

		$ret = $self->exec_dgmgrl("reinstate database $recoverdb");

		foreach (split(/\n/, $self->{'cmdout'})) {
			chomp;
			if (/ORA-\d+/) {
				#if database is not in open mode error then to avoid that 
				if ( /database not open/i ) {
					VCSAG_LOG_MSG("I","The reinstate database command returned oracle message ORA-01109: database is not open",1006);
				} elsif(/ORA-01031/ || /ORA-16653/ || /ORA-01017/) {
				        #if we receive "insufficient privileges error, try the command again.
					$self->{'orclerrmsg'} = $_;
					VCSAG_LOG_MSG("N", "The reinstate database command returned Oracle message $self->{'orclerrmsg'}. Will retry flashback recovery.", 545, $self->{'orclerrmsg'});
					$retryFlag = 1;	
					sleep(20);
				} else {
					$self->{'orclerrmsg'} = $_;
					VCSAG_LOG_MSG("E", "The reinstate database command returned Oracle message $self->{'orclerrmsg'}", 544, $self->{'orclerrmsg'});
					$ret = 1;
				}
				last;
			}
			last if (/SUCCESS/i);
		}
		$retryCnt--;
	} while ($retryCnt > 0 && $retryFlag);

	#If it was needed to restart remote db, clear sg faults
	$self->exec_action("CGclear");

	#If it was needed to restart remote db, probe remote oracle resource
	$self->exec_action("Resprobe");

	#unfreeze
	$self->exec_action("CGunfreeze");

	return $ret;
}


1;
